-------------------------------------------------------------------------------
--                L  O  C  A  L     V  A  R  I  A  B  L  E  S
-------------------------------------------------------------------------------

local addonName, L = ...;

local taxiFrameWidth = 0;

local flightNode = {};
local flightList = {};

-- Nodes to mark as cities
local nodeCities =
{
  [0x010A0199] = true, -- Stormwind
  [0x014801B5] = true, -- Shattrath
  [0x0194025B] = true, -- Lunarfall
  [0x019601D2] = true, -- Ironforge
  [0x01B601C1] = true, -- Thunder Bluff
  [0x01BF021D] = true, -- Shrine of Seven Stars
  [0x01F901EE] = true, -- Shrine of Two Moons
  [0x02030353] = true, -- Stormshield
  [0x02310275] = true, -- Orgrimmar
  [0x0269020A] = true, -- Dalaran
  [0x02690359] = true, -- Warspear
  [0x027401A0] = true, -- Undercity
  [0x02770152] = true, -- Frostwall Garrison
  [0x02E900CD] = true, -- The Exodar
  [0x0330024F] = true, -- Silvermoon City
  [0x03810156] = true, -- Darnassus
};

local nodeMapping =
{
  [L["Shattered Sun Staging Area"]] =
  {
    zone = L["Isle of Quel'Danas"],
    loc  = L["Shattered Sun Staging Area"]
  },
  [L["Dalaran"]] =
  {
    zone = L["Crystalsong Forest"],
    loc  = L["Dalaran"]
  },
  [L["Transitus Shield, Coldarra"]] =
  {
    zone = L["Borean Tundra"],
    loc  = L["Transitus Shield, Coldarra"]
  },
  [L["Hellfire Peninsula, The Dark Portal"]] =
  {
    zone = L["Hellfire Peninsula"],
    loc  = L["The Dark Portal"]
  },
  [L["Hellfire Peninsula, The Dark Portal, Alliance"]] =
  {
    zone = L["Hellfire Peninsula"],
    loc  = L["The Dark Portal"]
  },
  [L["Hellfire Peninsula, The Dark Portal, Horde"]] =
  {
    zone = L["Hellfire Peninsula"],
    loc  = L["The Dark Portal"]
  },
  [L["Acherus: The Ebon Hold"]] =
  {
    zone = L["Eastern Plaguelands"],
    loc  = L["Acherus: The Ebon Hold"]
  },
  [L["The Exodar"]] =
  {
    zone = L["Azuremyst Isle"],
    loc  = L["The Exodar"]
  },
  [L["Silvermoon City"]] =
  {
    zone = L["Eversong Woods"],
    loc  = L["Silvermoon City"]
  },
};

-------------------------------------------------------------------------------
--                L  O  C  A  L     F  U  N  C  T  I  O  N  S
-------------------------------------------------------------------------------

--
-- Function to create the list used in the scrolled frame. The list includes
-- elements for the zones.
--
local function flightListCreate()
  -- Local Variables
  local currentZone = "";
  local numNodes = #(flightNode);
  local searchText = NigglesFlights.searchBox:GetText();
  local addZone;
  local addLocations;
  local idx;
  local numMatches = 0;
  local striped;

  -- Check if there is any search text to match against
  if ((searchText ~= SEARCH) and
      (searchText ~= "") and
      (searchText ~= nil)) then
      searchText = string.lower(searchText);
  else
    searchText = nil;
  end

  -- Clear the list of flight nodes
  wipe(flightList);

  -- Process all the flight nodes
  nodeIdx = 1;
  while (nodeIdx <= numNodes) do
    -- Initialise some variables
    currentZone  = flightNode[nodeIdx].zone;
    addZone      = false;
    addLocations = false;

    -- Check if the zone should be added
    if (searchText ~= nil) then
      if (string.find(string.lower(currentZone), searchText,
           1, true) ~= nil) then
        addZone      = true;
        addLocations = true;
      else
        idx = nodeIdx;
        while ((idx <= numNodes) and
               (flightNode[idx].zone == currentZone)) do
          if (string.find(string.lower(flightNode[idx].location), searchText,
               1, true) ~= nil) then
            addZone = true;
            break;
          end
          idx = idx+1;
        end
      end
    else
      addZone      = true;
      addLocations = true;
    end

    -- Add the zone, if required
    if (addZone) then
      table.insert(flightList, -nodeIdx);
    end

    -- Add the locations, if required
    striped = false;
    while ((nodeIdx <= numNodes) and
           (flightNode[nodeIdx].zone == currentZone)) do
      if ((addLocations) or
          (string.find(string.lower(flightNode[nodeIdx].location),
            searchText, 1, true) ~= nil)) then
        -- Add the location to the flight list
        table.insert(flightList, nodeIdx);
        numMatches = numMatches+1;

        -- Determined if the location should be striped
        flightNode[nodeIdx].striped = striped;
        striped = not striped;
      end
      nodeIdx = nodeIdx+1;
    end
  end

  -- Set the scroll frame to the top
  HybridScrollFrame_SetOffset(NigglesFlights.scrollFrame, 0);
  NigglesFlightsUpdate();

  -- Show/Hide the filtered bar
  if (numMatches < numNodes) then
    NigglesFlights.filteredBar:SetFormattedText(L["FilteredShownHidden"],
      numMatches, numNodes-numMatches);
    NigglesFlights.filteredBar:SetHeight(18);
  else
    NigglesFlights.filteredBar:SetHeight(1);
  end

  return;
end

--
-- Function to compare 2 flight nodes
--
local function flightNodeSort(first, second)
  -- Local Variables
  local retValue;

  -- Work out which order the nodes should be
  if (first.zone ~= second.zone) then
    retValue = (naturalString(first.zone) < naturalString(second.zone));
  else
    retValue =
      (naturalString(first.location) < naturalString(second.location));
  end

  return retValue;
end


--
-- Function to show/hide the Niggles Flights frame
--
local function panelShow(show)
  -- Local Variables
  local button = NigglesFlightsToggle;
  local frame = NigglesFlights;

  -- Create the Niggles Flights frame, if required
  if ((frame == nil) and (show)) then
    -- Save the width of the Flights panel
    taxiFrameWidth = TaxiFrame:GetWidth();

    -- Create the Niggles Flights frame
    frame = CreateFrame("FRAME", "NigglesFlights", TaxiFrame,
      "NigglesFlightsTemplate");

    -- Initialise the filtered bar
    frame.filteredBar:Show();

    -- Initialise the scroll frame
    frame.scrollFrame.update = NigglesFlightsUpdate;
    HybridScrollFrame_CreateButtons(frame.scrollFrame,
      "NigglesFlightsButtonTemplate", 1, 0);
    frame.scrollFrame.scrollBar.doNotHide = true;
    frame.scrollFrame.scrollBar:SetPoint("TOPLEFT", frame,
      "TOPRIGHT", -25, -19);
  end

  -- Show/hide the Niggles Flights frame
  if (frame ~= nil) then
    if (show) then
      frame:Show();
      TaxiFrame:SetWidth(taxiFrameWidth+NigglesFlights:GetWidth());
      TaxiFrame.InsetBg:SetPoint("BOTTOMRIGHT", TaxiFrame.InsetBg:GetParent(),
        -(6+NigglesFlights:GetWidth()), 4);
    else
      frame:Hide();
      TaxiFrame:SetWidth(taxiFrameWidth);
      TaxiFrame.InsetBg:SetPoint("BOTTOMRIGHT", TaxiFrame.InsetBg:GetParent(),
        -6, 4);
    end
    SetUIPanelAttribute(TaxiFrame, "width", TaxiFrame:GetWidth());
    UpdateUIPanelPositions(TaxiFrame);
  end

  -- Set the texture of the toggle button
  if (button ~= nil) then
    if (show) then
      button:SetNormalTexture(
        "Interface\\Buttons\\UI-SpellbookIcon-PrevPage-Up");
      button:SetPushedTexture(
        "Interface\\Buttons\\UI-SpellbookIcon-PrevPage-Down");
      button:SetDisabledTexture(
        "Interface\\Buttons\\UI-SpellbookIcon-PrevPage-Disabled");
    else
      button:SetNormalTexture(
        "Interface\\Buttons\\UI-SpellbookIcon-NextPage-Up");
      button:SetPushedTexture(
        "Interface\\Buttons\\UI-SpellbookIcon-NextPage-Down");
      button:SetDisabledTexture(
        "Interface\\Buttons\\UI-SpellbookIcon-NextPage-Disabled");
    end
  end

  -- Save the panel's shown state
  NigglesDBC.settings.UiFlightListShown = show;

  return;
end

-------------------------------------------------------------------------------
--               G  L  O  B  A  L     F  U  N  C  T  I  O  N  S
-------------------------------------------------------------------------------

--
-- Function to process a button click in the Flights list
--
function NigglesFlightsButtonOnClick(button)
  -- Local variables
  local listIdx = button:GetID();

  -- Take a flight to the flight node clicked
  if (listIdx > 0) then
    TakeTaxiNode(flightNode[flightList[listIdx]].index);
    PlaySound("igMainMenuOptionCheckBoxOn");
  end

  return;
end

--
-- Function to handle OnEnter events for a button in the Flights List
--
function NigglesFlightsButtonOnEnter(button)
  -- Local variables
  local listIdx = button:GetID();

  -- Make the Taxi Frame show the route
  if ((listIdx > 0) and (flightList[listIdx] >=  0)) then
    TaxiNodeOnButtonEnter(
      _G["TaxiButton"..flightNode[flightList[listIdx]].index]);
  end

  return;
end

--
-- Function to enable/disable the Niggles Flights frame
--
function NigglesFlightsEnable()
  -- Enable or disable the Niggles Flights frame
  if (NigglesDBC.settings.UiFlightList == true) then
    -- Create the Niggles Flights toggle, if required
    if (NigglesFlightsToggle == nil) then
      CreateFrame("BUTTON", "NigglesFlightsToggle", TaxiFrame,
        "NigglesFlightsToggleTemplate");
    end
    NigglesFlightsToggle:Show();
    panelShow(NigglesDBC.settings.UiFlightListShown);
  else
    -- Hide the Niggles Flights frame
    panelShow(false);

    -- Hide the Niggles Flights toggle, if required
    if (NigglesFlightsToggle ~= nil) then
      NigglesFlightsToggle:Hide();
    end
  end

  return;
end

--
-- Function to update the Niggles Flights frame when it is shown
--
function NigglesFlightsOnShow()
  -- Local Variables
  local nodeType;
  local nodeName;
  local nodeX;
  local nodeY;
  local nodeCoord;
  local numNodes = NumTaxiNodes();
  local zoneName;
  local zoneLoc;

  -- Clear any filtering
  NigglesFlights.searchBox:SetText("");

  -- Clear the flight nodes
  wipe(flightNode);

  -- Get the info for all the nodes
  for nodeIdx = 1, numNodes do
    nodeType = TaxiNodeGetType(nodeIdx);
    if ((nodeType == "REACHABLE") or
        (nodeType == "UNREACHABLE") or
        (nodeType == "CURRENT")) then
      -- Work out the zone and location to use for the node
      nodeName = TaxiNodeName(nodeIdx);
      if (nodeMapping[nodeName] ~= nil) then
        zoneName = nodeMapping[nodeName].zone;
        zoneLoc  = nodeMapping[nodeName].loc;
      else
        zoneLoc, zoneName = string.match(nodeName, "(.*[^,]+), ([^,]*)");
        if ((zoneLoc == nil) or (zoneName == nil)) then
          zoneName = nodeName;
          zoneLoc  = nodeName;
        end
      end

      -- Work out a coordinate for the node
      nodeX, nodeY = TaxiNodePosition(nodeIdx);
      nodeCoords   = bit.lshift(nodeY*1000, 16)+math.floor(nodeX*1000);

      -- Add the node to the table
      table.insert(flightNode,
        {
          index    = nodeIdx,
          zone     = zoneName,
          location = zoneLoc,
          type     = nodeType,
          isCity   = (nodeCities[nodeCoords] == true);
        });
    end
  end

  -- Sort the table
  table.sort(flightNode, flightNodeSort);

  -- Create the flight list
  flightListCreate();

  return;
end

--
-- Function used to update the Flights list, based on the search text
--
function NigglesFlightsSearchUpdate(self, userInput)
  -- Update the search box's appearance
  SearchBoxTemplate_OnTextChanged(self);

  -- Create a new flight list, based on the search text
  flightListCreate();

  return;
end

--
-- Function to toggle the displaying of the extension to the Taxi panel
--
function NigglesFlightsToggleOnClick(button)
  -- Toggle the displaying of the Flights panel
  panelShow((NigglesFlights == nil) and true or (not NigglesFlights:IsShown()));
  PlaySound("igMainMenuOptionCheckBoxOn");

  return;
end

--
-- Function to update the Flights List in the Niggles Flights frame.
--
function NigglesFlightsUpdate()
  -- Local Variables
  local scrollFrame = NigglesFlights.scrollFrame;
  local button;
  local listIdx;
  local numRows = #(flightList);
  local numButtons = #scrollFrame.buttons;
  local topOffset = HybridScrollFrame_GetOffset(scrollFrame);
  local nodeIdx;
  local totalHeight;
  local displayedHeight;
  local fontColor;

  -- Make sure the top offset is valid
  if (topOffset > numRows) then
    topOffset = numRows;
    HybridScrollFrame_SetOffset(scrollFrame, topOffset);
  end

  -- Update all the buttons
  for idx = 1, numButtons do
    -- Get the current button
    button = scrollFrame.buttons[idx];

    -- Work out the index of the flight node to display in the button
    listIdx = topOffset+idx;
    if (listIdx <= numRows) then
      -- Set the button's ID
      button:SetID(listIdx);

      -- Check if the node is for a zone
      nodeIdx = flightList[listIdx];
      if (nodeIdx < 0) then
        button:SetText(flightNode[-nodeIdx].zone);
        fontColor = NORMAL_FONT_COLOR;
        button.zoneLeft:Show();
        button.zoneMiddle:Show();
        button.zoneRight:Show();
        button.stripe:Hide();
        button:Disable();
      else
        button:SetText(flightNode[nodeIdx].location);
        if (flightNode[nodeIdx].isCity) then
          fontColor = GREEN_FONT_COLOR;
        else
          fontColor = HIGHLIGHT_FONT_COLOR;
        end
        if (flightNode[nodeIdx].type == "REACHABLE") then
          button:Enable();
        else
          fontColor = GRAY_FONT_COLOR;
          button:Disable();
        end
        button.zoneLeft:Hide();
        button.zoneMiddle:Hide();
        button.zoneRight:Hide();
        button.stripe:SetShown(flightNode[nodeIdx].striped);
      end
      button.text:SetTextColor(fontColor.r, fontColor.g, fontColor.b, 1.0);

      -- Show the button
      button:Show();
    else
      -- Update the button
      button:Disable();
      button:Hide();
    end
  end

  -- Update the scroll frame's height
  totalHeight     = numRows*scrollFrame.buttons[1]:GetHeight();
  displayedHeight = numButtons*scrollFrame.buttons[1]:GetHeight();
  HybridScrollFrame_Update(scrollFrame, totalHeight, displayedHeight);

  return;
end
