-------------------------------------------------------------------------------
--                G  O  B  A  L     V  A  R  I  A  B  L  E  S
-------------------------------------------------------------------------------

local addonName, L = ...;

-------------------------------------------------------------------------------
--                L  O  C  A  L     V  A  R  I  A  B  L  E  S
-------------------------------------------------------------------------------

local MAX_TRADE_SKILLS = 6;

-- Mapping of spell ID to trade skill ID
local gatherSpellSkill =
{
  -- Archaeology
  [ 73979] = 794,
  -- Engineering
  [ 49383] = 202,
  -- Fishing
  [  7620] = 356,
  [  7731] = 356,
  [  7732] = 356,
  [ 18248] = 356,
  [ 33095] = 356,
  [ 51294] = 356,
  [ 88868] = 356,
  [110410] = 356,
  [131474] = 356,
  [131476] = 356,
  [131490] = 356,
  -- Herbalism
  [  2366] = 182,
  [  2368] = 182,
  [  3570] = 182,
  [ 11993] = 182,
  [ 28695] = 182,
  [ 50300] = 182,
  [ 74519] = 182,
  [110413] = 182,
  -- Mining
  [  2575] = 186,
  [  2576] = 186,
  [  3564] = 186,
  [ 10248] = 186,
  [ 29354] = 186,
  [ 50310] = 186,
  [ 74517] = 186,
  [102161] = 186,
  -- Skinning
  [  8613] = 393,
  [  8617] = 393,
  [  8618] = 393,
  [ 10768] = 393,
  [ 32678] = 393,
  [ 50305] = 393,
  [ 74522] = 393,
  [102216] = 393,
};

-- Info about the gathering skills
local gatherSkillInfo =
{
  [794] = -- Archaeology
  {
    index   = 0,
    tooltip = L["Requires Archaeology"]
  },
  [202] = -- Engineering
  {
    index   = 0,
    tooltip = UNIT_SKINNABLE_BOLTS
  },
  [356] = -- Fishing
  {
    index   = 0,
    tooltip = ""
  },
  [182] = -- Herbalism
  {
    index   = 0,
    tooltip = UNIT_SKINNABLE_HERB
  },
  [186] = -- Mining
  {
    index   = 0,
    tooltip = UNIT_SKINNABLE_ROCK
  },
  [393] = -- Skinning
  {
    index   = 0,
    tooltip = UNIT_SKINNABLE_LEATHER
  },
};

local eventFrame;

--
-- These value have been copied from 'Blizzard_TradeSkillUI.lua' because the
-- Trade Skill panel is implemented as an on-demand addon.
--
local tradeSkillTypeInfo =
{
  ["trivial"] = {r = 0.50, g = 0.50, b = 0.50},
  ["easy"]    = {r = 0.25, g = 0.75, b = 0.25},
  ["medium"]  = {r = 1.00, g = 1.00, b = 0.00},
  ["optimal"] = {r = 1.00, g = 0.50, b = 0.25},
};

local tradeSkillInfo =
{
  professionIdx  = 0,
  recipeIdx      = 0,
  maxRank        = 0,
  numProfessions = 0,
  numMaxed       = 0,
};

local recipeInfo =
{
  index     = 0,
  spellId   = 0,
  skillType = "",
  rank      = 0,
  maxRank   = 0,
  skillName = "",
};

local tradeSkillFrameInfo =
{
  buttons      = {},
  textures     = {},
  searchPoints = {},
  searchWidth  = 0,
};

local tradeSkillHistory = {};
local tradeSkillIndices = {};
local tradeSkillHistoryEnabled = true;

-------------------------------------------------------------------------------
--             L  O  C  A  L     D  E  F  I  N  I  T  I  O  N  S
-------------------------------------------------------------------------------

local tradeSkillIndicesIsValid;
local tradeSkillIndicesUpdate;
local tradeSkillIndicesUpdateRestore;
local tradeSkillClearFilters;
local tradeSkillFrameOnShow;
local tradeSkillHistoryBtnEnable;
local tradeSkillHistoryBtnOnClick;
local tradeSkillRecipeSelect;
local tradeSkillTypeFromTooltip;

-------------------------------------------------------------------------------
--                L  O  C  A  L     F  U  N  C  T  I  O  N  S
-------------------------------------------------------------------------------

--
-- Function called when a recipe reagent is clicked. It checks if the reagent
-- can be created by the current trade skill and is known by the current
-- character. If it is, the recipe is selected.
--
local function recipeReagentBtnOnClick(self, mouseButton)
  -- Local Variables
  local recipeIdx = (TradeSkillFrame.selectedSkill or 0);
  local recipeName = GetTradeSkillInfo(recipeIdx);
  local reagentName;

  -- Check if the button click should be processed
  if ((mouseButton == "LeftButton") and
      (not IsModifierKeyDown()) and
      (recipeIdx > 0)) then
    -- Get the reagent's name
    reagentName = "I|"..
      (GetTradeSkillReagentInfo(recipeIdx, self:GetID()) or "");
    if (tradeSkillIndices[reagentName] ~= nil) then
      -- Clear all filtering
      tradeSkillHistoryEnabled = false;
      tradeSkillClearFilters();
      TradeSkillUpdateFilterBar();
      tradeSkillHistoryEnabled = true;

      -- Check the reagent's index is valid
      if (tradeSkillIndicesIsValid(reagentName)) then
        -- Check if the reagent has a single recipe
        if ((tradeSkillIndices[reagentName] > 0) and
            (tradeSkillIndices[reagentName] <= GetNumTradeSkills())) then
          tradeSkillRecipeSelect(tradeSkillIndices[reagentName]);
        elseif (tradeSkillIndices[reagentName] == -1) then
          -- Set the reagent as the search string
          TradeSkillFrameSearchBox:SetText(reagentName:sub(3));
          tradeSkillHistory[tradeSkillInfo.professionIdx]:Add(reagentName);
          tradeSkillHistoryBtnEnable(tradeSkillInfo.professionIdx);
        end
      end
    end
  end

  return;
end

--
-- Function hooked to 'TradeSkillFrame_SetSelection'. It shows/hides the
-- textures used to indicate recipe reagents that can be made by the current
-- trade skill and character as well as adding the selected recipe to the
-- history, if required.
--
local function recipeOnSelection()
  -- Local Variables
  local button;
  local professionIdx = tradeSkillInfo.professionIdx;
  local reagentName;
  local recipeIdx = (TradeSkillFrame.selectedSkill or 0);
  local recipeName;
  local textures = tradeSkillFrameInfo.textures;

  -- Check there is a recipe displayed
  if ((TradeSkillFrame:IsShown()) and (recipeIdx > 0)) then
    -- Show/Hide the textures that indicate a reagent can be created
    for idx = 1, GetTradeSkillNumReagents(recipeIdx) do
      button = _G["TradeSkillReagent"..idx]
      if ((button ~= nil) and
          (button:IsShown()) and
          (textures[idx] ~= nil)) then
        reagentName = GetTradeSkillReagentInfo(recipeIdx, idx);
        textures[idx]:SetShown(((reagentName ~= nil) and
          (tradeSkillIndices["I|"..reagentName] ~= nil)));
      end
    end

    -- Add the recipe to the history, if required
    if ((professionIdx > 0) and
        (recipeIdx >= 1) and
        (recipeIdx <= GetNumTradeSkills()) and
        (tradeSkillHistoryEnabled)) then
      recipeName = GetTradeSkillInfo(recipeIdx);
      if ((recipeName ~= nil) and
          (tradeSkillHistory[professionIdx]:Get() ~= "R|"..recipeName)) then
        tradeSkillHistory[professionIdx]:Add("R|"..recipeName);
        tradeSkillHistoryBtnEnable(professionIdx);
      end
    end
  end

  return;
end

--
-- Function to clear all filters for the Trade Skill panel
--
tradeSkillClearFilters = function()
  -- Local Variables
  local filterTbl = TradeSkillFrame.filterTbl;

  -- Clear all trade skill filters
  filterTbl.hasMaterials  = false;
  filterTbl.hasSkillUp    = false;
  filterTbl.subClassValue = -1;
  filterTbl.slotValue     = -1;
  TradeSkillFrameSearchBox:ClearFocus();
  TradeSkillFrameSearchBox:SetText("");
  TradeSkillFrameSearchBox:GetScript("OnEditFocusLost")(
    TradeSkillFrameSearchBox);
  SetTradeSkillItemNameFilter();
  TradeSkillOnlyShowSkillUps(filterTbl.hasSkillUp);
  TradeSkillOnlyShowMakeable(filterTbl.hasMaterials);
  SetTradeSkillCategoryFilter(filterTbl.subClassValue, 0);
  SetTradeSkillInvSlotFilter(filterTbl.slotValue, 1, 1);
  ExpandTradeSkillSubClass(0);
  CloseDropDownMenus();

  return;
end

--
-- Function hooked to 'DoTradeSkill' to capture the spell ID of the item being
-- created so that its skill type can be obtained for each iteration
--
local function tradeSkillDoRecipe(idx, count)
  -- Local Variables
  local spellId;

  -- Extract the spell ID from the recipe's link
  spellId = string.gsub(GetTradeSkillRecipeLink(idx),
    "^[^:]*:([^|]*)\|h.*$", "%1");
  if (spellId ~= nil) then
    -- Save the spell ID
    recipeInfo.index     = tonumber(idx);
    recipeInfo.spellId   = tonumber(spellId);
    recipeInfo.skillName = GetTradeSkillLine();
  else
    recipeInfo.index     = 0;
    recipeInfo.spellId   = 0;
    recipeInfo.skillName = "";
  end

  return;
end

--
-- Function called when the Trade Skill frame is shown. It enables/disables
-- Niggles features, depending on the current settings.
--
tradeSkillFrameOnShow = function(self)
  -- Local Variables
  local button;
  local idx = 1;
  local itemName;
  local texturePath;
  local textures = tradeSkillFrameInfo.textures;

  -- Set up the recipe history buttons
  for idx, suffix in ipairs({"PrevPage", "NextPage"}) do
    -- Create the button, if required
    if ((NigglesDBC.settings.ProfGeneralHistory) and
        (tradeSkillFrameInfo.buttons[idx] == nil)) then
      -- Work out the path for the button's textures
      texturePath = "Interface\\Buttons\\UI-SpellbookIcon-"..suffix;

      -- Create the button
      button = CreateFrame("Button", nil, TradeSkillFrame);
      button:SetWidth(23);
      button:SetHeight(22);
      button:SetNormalTexture(texturePath.."-Up");
      button:SetPushedTexture(texturePath.."-Down");
      button:SetDisabledTexture(texturePath.."-Disabled");
      button:SetHighlightTexture(
        "Interface\\Buttons\\UI-Common-MouseHilight", "ADD");
      button:SetPoint("TOPLEFT", TradeSkillRankFrame, "BOTTOMLEFT",
        5+((idx-1)*20), -8);
      button:SetID(idx);
      button:SetScript("OnClick", tradeSkillHistoryBtnOnClick);
      button:Disable();
      tradeSkillFrameInfo.buttons[idx] = button;
    else
      button = tradeSkillFrameInfo.buttons[idx];
    end

    -- Show/Hide the button
    if (button ~= nil) then
      button:SetShown(NigglesDBC.settings.ProfGeneralHistory);
    end
  end

  -- Adjust the position of the search box
  if (NigglesDBC.settings.ProfGeneralHistory) then
    TradeSkillFrameSearchBox:SetPoint("TOPLEFT",
      tradeSkillFrameInfo.buttons[2], "TOPRIGHT", 9, -1);
    TradeSkillFrameSearchBox:SetWidth(131);
  else
    NigglesFramePointsSet(TradeSkillFrameSearchBox,
      tradeSkillFrameInfo.searchPoints);
    TradeSkillFrameSearchBox:SetWidth(tradeSkillFrameInfo.searchWidth);
  end

  -- Create the textures that indicate creatable recipe reagents, if required
  if (NigglesDBC.settings.ProfGeneralReagents) then
    while (_G["TradeSkillReagent"..idx] ~= nil) do
      button = _G["TradeSkillReagent"..idx];
      if (textures[idx] == nil) then
        textures[idx] =
          _G["TradeSkillReagent"..idx]:CreateTexture(nil, "OVERLAY");
        textures[idx]:SetTexture("Interface\\SPELLBOOK\\RotationIconFrame");
        textures[idx]:SetPoint("TOPLEFT", button.Icon, -6, 6);
        textures[idx]:SetPoint("BOTTOMRIGHT", button.Icon, 7, -7);
      end
      idx = idx+1;
    end
  end

  return;
end

--
-- Function to enable/disable the trade skill history buttons
--
tradeSkillHistoryBtnEnable = function(tradeSkillIdx)
  -- Enable/disable the 'Back' button
  if (tradeSkillFrameInfo.buttons[1] ~= nil) then
    tradeSkillFrameInfo.buttons[1]:SetEnabled(
      (tradeSkillIdx ~= nil) and
      (tradeSkillHistory[tradeSkillIdx] ~= nil) and
      (tradeSkillHistory[tradeSkillIdx]:IsMoveValid(true)));
  end

  -- Enable/disable the 'Forward' button
  if (tradeSkillFrameInfo.buttons[2] ~= nil) then
    tradeSkillFrameInfo.buttons[2]:SetEnabled(
      (tradeSkillIdx ~= nil) and
      (tradeSkillHistory[tradeSkillIdx] ~= nil) and
      (tradeSkillHistory[tradeSkillIdx]:IsMoveValid(false)));
  end

  return;
end

--
-- Function called when a Trade skill History button is clicked.
--
tradeSkillHistoryBtnOnClick = function(self, mouseButton)
  -- Local Variables
  local backward = (self:GetID() == 1);
  local history = tradeSkillHistory[tradeSkillInfo.professionIdx];
  local historyIdx = 0;
  local historyName;
  local professionIdx = tradeSkillInfo.professionIdx;

  -- Check moving in the specified direction is valid
  if (history:IsMoveValid(backward)) then
    -- Disable updates to the trade skill history
    tradeSkillHistoryEnabled = false;

    -- Clear all filtering
    tradeSkillClearFilters();
    TradeSkillUpdateFilterBar();

    -- Get the name of the recipe in the appropriate direction
    historyName = history:Move(backward);

    -- Check the recipe is valid
    if ((tradeSkillIndicesIsValid(historyName)) and
        (tradeSkillIndices[historyName] ~= nil)) then
      -- Select the recipe, or search for the recipe item
      historyIdx = tradeSkillIndices[historyName];
      if ((historyIdx > 0) and (historyIdx <= GetNumTradeSkills())) then
        -- Select the recipe
        tradeSkillRecipeSelect(historyIdx);
      elseif (historyIdx == -1) then
        -- Set the recipe item as the search string
        TradeSkillFrameSearchBox:SetText(historyName:sub(3));
      end
    end
    TradeSkillFrame_Update();

    -- Enable updates to the trade skill history
    tradeSkillHistoryEnabled = true;
  end

  -- Update the state of the trade skill history buttons
  tradeSkillHistoryBtnEnable(tradeSkillInfo.professionIdx);

  return;
end

--
-- Function to make sure there is a valid index in the 'tradeSkillIndices'
-- list for a recipe or item. It checks that the item created by the recipe
-- matches the specified item name. This function is defensive, to deal with
-- situations where the trade skill indices list has become corrupted for some
-- reason, usually because it was built before the trade skill is truly ready.
--
tradeSkillIndicesIsValid = function(recipeItemName)
  -- Local Variables
  local tradeSkillIdx;
  local isValid = false;
  local indexType;

  -- Check an index can be found for the recipe/item
  for idx = 1, 2 do
    -- Get the index for the recipe/item
    tradeSkillIdx = tradeSkillIndices[recipeItemName];

    -- Check if only one recipe is known for the recipe/item
    if ((tradeSkillIdx ~= nil) and (tradeSkillIdx > 0)) then
      indexType      = recipeItemName:sub(1, 1);
      recipeItemName = recipeItemName:sub(3);
      if (indexType == "R") then
        isValid = (GetTradeSkillInfo(tradeSkillIdx) == recipeItemName);
        break;
      elseif (indexType == "I") then
        isValid = (string.match((GetTradeSkillItemLink(tradeSkillIdx) or ""),
          "|h%[(.*)%]|h|r") == recipeItemName);
        break;
      end
    elseif (tradeSkillIdx == -1) then
      isValid = true;
      break;
    end

    -- Update the trade skill indices, because they appear to be invalid
    if (idx == 1) then
      tradeSkillIndicesUpdate(false)
    end
  end

  return isValid;
end

--
-- Function to update the list of indices for recipes and the items they
-- create for the current trade skill
--
tradeSkillIndicesUpdate = function(restoreFilters)
  -- Local Variables
  local itemName;
  local itemNameFilter = GetTradeSkillItemNameFilter();
  local recipeName;

  -- Check if the trade skill isn't ready to be interrogated
  if (not IsTradeSkillReady()) then
    -- Call the function again in a short while
    C_Timer.After(0.25,
      (restoreFilters and tradeSkillIndicesUpdateRestore or
        tradeSkillIndicesUpdate));
    return false;
  end

  -- Clear the current filter settings
  tradeSkillClearFilters();

  -- Create a list of indices for the trade skill's recipes and their items
  wipe(tradeSkillIndices);
  for idx = 1, GetNumTradeSkills() do
    -- Add the name of the recipe to the list
    recipeName = GetTradeSkillInfo(idx);
    if (recipeName ~= nil) then
      -- Prepend the recipe name to indicate it is a recipe
      recipeName = "R|"..recipeName;

      -- Check the name hasn't already been seen
      if (tradeSkillIndices[recipeName] == nil) then
        -- Save the recipe's name and index
        tradeSkillIndices[recipeName] = idx;
      elseif (idx ~= tradeSkillIndices[recipeName]) then
        -- Set to a special value to indicate multiple entries
        tradeSkillIndices[recipeName] = -1;
      end
    end

    -- Get the name of the item created by the recipe
    itemName = string.match((GetTradeSkillItemLink(idx) or ""),
      "|h%[(.*)%]|h|r");
    if (itemName ~= nil) then
      -- Prepend the item name to indicate it is a recipe
      itemName = "I|"..itemName;

      -- Check the name hasn't already been seen
      if (tradeSkillIndices[itemName] == nil) then
        -- Save the item's name and the recipe's index
        tradeSkillIndices[itemName] = idx;
      elseif (idx ~= tradeSkillIndices[itemName]) then
        -- Set to a special value to indicate multiple entries
        tradeSkillIndices[itemName] = -1;
      end
    end
  end

  -- Restore the filter settings, if required
  if (restoreFilters == true) then
    TradeSkillFrameSearchBox:SetText(itemNameFilter or "");
    TradeSkillFrameSearchBox:GetScript("OnEditFocusLost")(
      TradeSkillFrameSearchBox);
    TradeSkillOnlyShowSkillUps(TradeSkillFrame.filterTbl.hasSkillUp);
    TradeSkillOnlyShowMakeable(TradeSkillFrame.filterTbl.hasMaterials);
    SetTradeSkillCategoryFilter(TradeSkillFrame.filterTbl.subClassValue, 0);
    SetTradeSkillInvSlotFilter(TradeSkillFrame.filterTbl.slotValue, 1, 1);
  end
  TradeSkillUpdateFilterBar();

  return true;
end

--
-- Function used as a wrapper around 'tradeSkillIndicesUpdate' to allow it to
-- be used as a callback for 'C_Timer.After'.
--
tradeSkillIndicesUpdateRestore = function()
  tradeSkillIndicesUpdate(true);
end

--
-- Function to process events
--
local function tradeSkillOnEvent(self, event, ...)
  -- Local Variables
  local _;
  local skillId;

  -- Process the event
  if ((event == "UNIT_SPELLCAST_START") or
      (event == "UNIT_SPELLCAST_CHANNEL_START")) then
    -- Local Variables
    local skillId;

    -- Get the spell ID
    local _, _, _, _, spellId = ...;
    skillId = gatherSpellSkill[spellId];

    -- Check if a gathering skill is being cast
    if ((skillId ~= nil) and (gatherSkillInfo[skillId] ~= nil)) then
      -- Get the skill type from the tooltip
      recipeInfo.skillType = tradeSkillTypeFromTooltip(skillId);

      -- Get the profession's current and max ranks
      recipeInfo.skillName, _, recipeInfo.rank, recipeInfo.maxRank =
        GetProfessionInfo(gatherSkillInfo[skillId].index);
    elseif (spellId == recipeInfo.spellId) then
      -- Get the skill type
      if (recipeInfo.index > 0) then
        -- Get the skill type
        _, recipeInfo.skillType = GetTradeSkillInfo(recipeInfo.index);
        _, recipeInfo.rank, recipeInfo.maxRank = GetTradeSkillLine();
      else
        recipeInfo.skillType = "optimal";
      end
    else
      recipeInfo.skillType = "";
    end

    -- Check if a warning should be displayed
    if ((not IsNPCCrafting()) and
        (recipeInfo.skillType ~= "") and
        (recipeInfo.skillType ~= "trivial") and
        (recipeInfo.maxRank < tradeSkillInfo.maxRank) and
        (recipeInfo.rank >= recipeInfo.maxRank-5)) then
      if (recipeInfo.rank >= recipeInfo.maxRank) then
        Niggles:AddWarning(L["Trade Skill Capped"], recipeInfo.skillName);
      else
        Niggles:AddWarning(L["Trade Skill Near Cap"], recipeInfo.skillName);
      end
      PlaySound("RaidWarning");
    end
  elseif (event == "SKILL_LINES_CHANGED") then
    -- Initialise some variables
    local skillRank;
    local professions = {};
    tradeSkillInfo.numProfessions = 0;
    tradeSkillInfo.numMaxed = 0;

    -- Get the character's professions
    professions[1], professions[2], professions[3],
    professions[4], professions[5], professions[6] = GetProfessions();

    -- Get the index of each gather professions the character knows
    for _, professionIdx in pairs(professions) do
      -- Get info for the profession
      _, _, skillRank, _, _, _, skillId = GetProfessionInfo(professionIdx);

      -- Check if the skill is a gather skill
      if (gatherSkillInfo[skillId] ~= nil) then
        -- Save the profession's skill index
        gatherSkillInfo[skillId].index = professionIdx;
      end

      -- Check if the skill is at the maximum rank possible
      tradeSkillInfo.numProfessions = tradeSkillInfo.numProfessions+1;
      if (skillRank == tradeSkillInfo.maxRank) then
        tradeSkillInfo.numMaxed = tradeSkillInfo.numMaxed+1;
      end
    end

    -- Check if there are any professions that aren't maxed
    if (tradeSkillInfo.numMaxed < tradeSkillInfo.numProfessions) then
      eventFrame:RegisterUnitEvent("UNIT_SPELLCAST_START", "player");
      eventFrame:RegisterUnitEvent("UNIT_SPELLCAST_CHANNEL_START", "player");
    else
      eventFrame:UnregisterEvent("UNIT_SPELLCAST_START");
      eventFrame:UnregisterEvent("UNIT_SPELLCAST_CHANNEL_START");
    end
  elseif (event == "TRADE_SKILL_SHOW") then
    -- Local Variables
    local tradeSkillName = GetTradeSkillLine();

    -- Check a known profession is being shown
    if (tradeSkillName ~= "UNKNOWN") then
      -- Work out the index of the profession being shown
      if ((IsTradeSkillGuild()) or
          (IsTradeSkillLinked()) or
          (IsNPCCrafting())) then
        -- Use a special history list for other characters
        tradeSkillInfo.professionIdx = #tradeSkillHistory;
        tradeSkillHistory[tradeSkillInfo.professionIdx]:Clear();
      else
        -- Get the ID of all the character's professions
        local professions = {};
        professions[1], professions[2], professions[3],
        professions[4], professions[5], professions[6] = GetProfessions();

        -- Work out the index of the current profession
        tradeSkillInfo.professionIdx = 0;
        for idx, id in pairs(professions) do
          if (GetProfessionInfo(id) == tradeSkillName) then
            tradeSkillInfo.professionIdx = idx;
            break;
          end
        end
      end

      -- Update the list of items that can be created by known recipes
      tradeSkillIndicesUpdate(true);
    else
      tradeSkillInfo.professionIdx = 0;
    end

    -- Enable/disable the trade skill history buttons
    tradeSkillHistoryBtnEnable(tradeSkillInfo.professionIdx);
  elseif (event == "SPELLS_CHANGED") then
    if (GetTradeSkillLine() ~= "UNKNOWN") then
      tradeSkillIndicesUpdate(true);
    end
  elseif (event == "TRADE_SKILL_HIDE") then
    wipe(tradeSkillIndices);
  end

  return;
end

--
-- Function to scroll the list of recipes so the specified recipe is visible
--
tradeSkillRecipeSelect = function(recipeIdx)
  -- Local Variables
  local bottomIdx;
  local scrollBar = TradeSkillListScrollFrameScrollBar;
  local topIdx;

  -- Initialise some variables
  topIdx    =
    (TradeSkillListScrollFrameScrollBar:GetValue()/TRADE_SKILL_HEIGHT)+1;
  bottomIdx = topIdx+TRADE_SKILLS_DISPLAYED-1;

  -- Scroll the recipe into view, if required
  if (recipeIdx == topIdx-1) then
    scrollBar:SetValue(TRADE_SKILL_HEIGHT*(topIdx-2));
  elseif (recipeIdx == bottomIdx+1) then
    scrollBar:SetValue(TRADE_SKILL_HEIGHT*topIdx);
  elseif ((recipeIdx < topIdx) or (recipeIdx > bottomIdx)) then
    scrollBar:SetValue(TRADE_SKILL_HEIGHT*
      (math.min(math.max(recipeIdx-math.floor(TRADE_SKILLS_DISPLAYED/2), 1),
        (GetNumTradeSkills()-TRADE_SKILLS_DISPLAYED+1))-1));
  else
    scrollBar:SetValue(TRADE_SKILL_HEIGHT*(topIdx-1));
  end

  -- Select the specified recipe
  TradeSkillFrame_SetSelection(recipeIdx);
  TradeSkillFrame_Update();

  return;
end

--
-- Function hooked to 'SolveArtifact' to capture the spell ID of the artifact
-- being solved.
--
local function tradeSkillSolveArtifact(...)
  -- Local Variables
  local _;
  local spellId = NigglesArchaeologyGetSpellId();

  -- Check a spell ID could be found for the selected artifact
  if (spellId > 0) then
    recipeInfo.index     = 0;
    recipeInfo.spellId   = spellId;
    recipeInfo.skillType = "optimal";
    recipeInfo.skillName, _, recipeInfo.rank, recipeInfo.maxRank =
      GetProfessionInfo(gatherSkillInfo[794].index);
  end

  return;
end

--
-- Function to get the skill type from the GameTooltip when a world object is
-- gathered.
--
tradeSkillTypeFromTooltip = function(skillId)
  -- Local Variables
  local colourR;
  local colourG;
  local colourB;
  local skillType = "optimal";

  -- Process each line of the Game Tooltip
  for idx = GameTooltip:NumLines(), 1, -1 do
    -- Check if the tooltip line matches the gather skill's tooltip
    if (_G["GameTooltipTextLeft"..idx]:GetText() ==
         gatherSkillInfo[skillId].tooltip) then
      -- Get the colour of the tooltip line
      colourR, colourG, colourB = _G["GameTooltipTextLeft"..idx]:GetTextColor();

      -- Find the skill type, based on the colour
      for currentType, currentInfo in pairs(tradeSkillTypeInfo) do
        if ((math.abs(colourR-currentInfo.r) < 0.01) and
            (math.abs(colourG-currentInfo.g) < 0.01) and
            (math.abs(colourB-currentInfo.b) < 0.01)) then
          skillType = currentType;
        end
      end
      break;
    end
  end

  return skillType;
end

-------------------------------------------------------------------------------
--               G  L  O  B  A  L     F  U  N  C  T  I  O  N  S
-------------------------------------------------------------------------------

--
-- Function to enable/disable the Niggles trade skill features
--
function NigglesTradeSkillEnable()
  -- Local Variables
  local button;
  local idx = 1;

  -- Check if the frame used to monitor events has already be created
  if (eventFrame == nil) then
    -- Create the event frame
    eventFrame = CreateFrame("FRAME", nil);
    eventFrame:SetScript("OnEvent", tradeSkillOnEvent);

    -- Initialise some info
    tradeSkillInfo.maxRank = PROFESSION_RANKS[#PROFESSION_RANKS][1];

    -- Hook the functions that use trade skill recipes
    hooksecurefunc("DoTradeSkill",  tradeSkillDoRecipe);
    hooksecurefunc("SolveArtifact", tradeSkillSolveArtifact);
  end

  -- (Un)register for events, depending on the current settings
  if (NigglesDBC.settings.ProfGeneralReagents) then
    eventFrame:RegisterEvent("TRADE_SKILL_SHOW");
    eventFrame:RegisterEvent("TRADE_SKILL_HIDE");
    eventFrame:RegisterEvent("SPELLS_CHANGED");
  else
    eventFrame:UnregisterEvent("TRADE_SKILL_SHOW");
    eventFrame:UnregisterEvent("TRADE_SKILL_HIDE");
    eventFrame:UnregisterEvent("SPELLS_CHANGED");
  end
  if (NigglesDBC.settings.WarningTradeSkillCap) then
    eventFrame:RegisterEvent("SKILL_LINES_CHANGED");
    tradeSkillOnEvent(eventFrame, "SKILL_LINES_CHANGED");
  else
    eventFrame:UnregisterEvent("UNIT_SPELLCAST_START");
    eventFrame:UnregisterEvent("UNIT_SPELLCAST_CHANNEL_START");
    eventFrame:UnregisterEvent("SKILL_LINES_CHANGED");
  end

  -- Update the Trade Skill frame, based on current settings
  if ((TradeSkillFrame ~= nil) and
      (TradeSkillFrame:IsShown())) then
    tradeSkillFrameOnShow(TradeSkillFrame);
  end

  -- Update/Clear the recipe list, based on current settings
  if ((NigglesDBC.settings.ProfGeneralReagents) and
      (GetTradeSkillLine() ~= "UNKNOWN")) then
    tradeSkillIndicesUpdate(true);
  else
    -- Clear any trade skill indices
    wipe(tradeSkillIndices);
  end

  return;
end

--
-- Function called when Blizzard's Trade Skill addon is loaded
--
function NigglesTradeSkillOnLoad()
  -- Local Variables
  local button;
  local texture;
  local idx = 1;

  -- Hook the Trade Skill Frame
  TradeSkillFrame:HookScript("OnShow", tradeSkillFrameOnShow);

  -- Hook Blizzard functions
  hooksecurefunc("TradeSkillFrame_SetSelection", recipeOnSelection);

  -- Hook the 'OnClick' script handler for the reagent buttons
  while (_G["TradeSkillReagent"..idx] ~= nil) do
    button = _G["TradeSkillReagent"..idx];
    button:HookScript("OnClick", recipeReagentBtnOnClick);
    idx = idx+1;
  end

  -- Save the points and width of the Trade Skill frame's search box
  NigglesFramePointsGet(TradeSkillFrameSearchBox,
    tradeSkillFrameInfo.searchPoints);
  tradeSkillFrameInfo.searchWidth = TradeSkillFrameSearchBox:GetWidth();

  -- Initialise the history for each trade skill
  for idx = 1, MAX_TRADE_SKILLS+1 do
    tradeSkillHistory[idx] = NigglesHistoryClass:New();
  end

  -- Clear all filtering
  tradeSkillClearFilters();

  return;
end
